# IProxy.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import abstractmethod
from typing import Any

from .INotifier import INotifier


class IProxy(INotifier):
    """
    The interface definition for a PureMVC Proxy.

    In PureMVC, 'IProxy' implementors assume these responsibilities:</P>

    Implement a common method which returns the name of the Proxy.
    Provide methods for setting and getting the data object.

    Additionally, 'IProxy's typically:

    Maintain references to one or more pieces of model data.
    Provide methods for manipulating that data.
    Generate 'INotifications' when their model data changes.
    Expose their name as a 'public static const' called 'NAME',
    if they are not instantiated multiple times.
    Encapsulate interaction with local or remote services used to fetch and persist model data.
    """

    @property
    @abstractmethod
    def proxy_name(self) -> str:
        """
        Get the Proxy name

        :return: The name of the proxy as a string.
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def data(self) -> Any:
        """
        Get the data object

        :return: The data of the method.
        :rtype: Any
        """
        pass

    @data.setter
    def data(self, value: Any):
        """
        Get the data object

        :param value: The value to be set for the `data` attribute.
        :type value: Any
        :return: None
        """
        pass

    @abstractmethod
    def on_register(self):
        """Called by the Model when the Proxy is registered."""
        pass

    @abstractmethod
    def on_remove(self):
        """Called by the Model when the Proxy is removed."""
        pass



# IController.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import ABC, abstractmethod
from typing import Callable

from .ICommand import ICommand
from .INotification import INotification


class IController(ABC):
    """
    The interface definition for a PureMVC Controller.

    In PureMVC, an `IController` implementor follows the 'Command and Controller' strategy, and
    assumes these responsibilities:

    Remembering which `ICommand`s are intended to handle which `INotifications`.

    Registering itself as an `IObserver` with the `View` for each `INotification`
    that it has an `ICommand` mapping for.

    Creating a new instance of the proper `ICommand` to handle a given `INotification` when notified by the `View`.

    Calling the `ICommand`'s `execute` method, passing in the `INotification`.

    See Also
    --------
    :class:`puremvc.interfaces.INotification`
    :class:`puremvc.interfaces.ICommand`
    """

    @abstractmethod
    def register_command(self, notification_name: str, factory: Callable[[], ICommand]):
        """
        Register a particular `ICommand` class as the handler for a particular `INotification`.

        :param notification_name: The name of the `INotification`
        :type notification_name: str
        :param factory: A factory function that returns an instance of ICommand.
        :type factory: Callable[[], ICommand]
        :return: None
        """
        pass

    @abstractmethod
    def execute_command(self, notification: INotification):
        """
        Execute the `ICommand` previously registered as the handler for
        `INotification` with the given notification name.

        :param notification: The `INotification` to execute the associated `ICommand` for
        :type notification: INotification
        :return: None
        """
        pass

    @abstractmethod
    def has_command(self, notification_name: str) -> bool:
        """
        Check if a Command is registered for a given Notification.

        :param notification_name: The name of the `INotification`
        :type notification_name: str
        :return: True if a command is registered for the notification_name, False otherwise.
        :rtype: bool
        """
        pass

    @abstractmethod
    def remove_command(self, notification_name: str):
        """
        Remove a previously registered `ICommand` to `INotification` mapping.

        :param notification_name: The name of the `INotification` to remove the `ICommand` mapping for
        :type notification_name: str
        :return: None
        """
        pass



# Facade.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

import threading
from typing import Dict, Callable, Any

from puremvc.core import Controller, Model, View
from puremvc.interfaces import IFacade, INotification, ICommand, IProxy, IMediator, IController, IModel, IView
from puremvc.patterns.observer import Notification


class Facade(IFacade):
    """
    A base Multiton `IFacade` implementation.

    In PureMVC, the `Facade` class assumes these responsibilities:

    Initializing the `Model`, `View` and `Controller` Singletons.

    Providing all the methods defined by the `IModel, IView, & IController` interfaces.

    Providing the ability to override the specific `Model`, `View` and `Controller` Singletons created.

    Providing a single point of contact to the application for registering `Commands` and notifying `Observers`

    See Also
    --------
    :class:`puremvc.core.Model`
    :class:`puremvc.core.View`
    :class:`puremvc.core.Controller`
    :class:`puremvc.patterns.observer.Notification`
    :class:`puremvc.patterns.mediator.Mediator`
    :class:`puremvc.patterns.proxy.Proxy`
    :class:`puremvc.patterns.command.SimpleCommand`
    :class:`puremvc.patterns.command.MacroCommand`
    """
    instanceMap: Dict[str, IFacade] = dict()
    instanceMapLock = threading.Lock()

    """MULTITON_MSG (str): Multiton error message"""
    MULTITON_MSG = "Facade instance for this Multiton key already constructed!"

    def __init__(self, key: str):
        """
        Constructor.

        This `IFacade` implementation is a Multiton, so you should not call
        the constructor directly, but instead call the static Factory method,
        passing the unique key for this instance `Facade.get_instance(multitonKey, lambda k: Facade(k))`

        :param key: The Multiton key for the Facade instance.
        :type key: str
        :raises Exception: If a Facade instance has already been constructed with the same Multiton key.
        """
        if Facade.instanceMap.get(key) is not None:
            raise Exception(Facade.MULTITON_MSG)
        self.multitonKey = None
        self.controller = None
        self.model = None
        self.view = None
        self.initialize_notifier(key)
        Facade.instanceMap[key] = self
        self.initialize_facade()

    def initialize_facade(self):
        """
        Initialize the Multiton `Facade` instance.

        Called automatically by the constructor. Override in your subclass to
        do any subclass specific initializations. Be sure to call
        `super.initialize_facade()`, though.

        :return: None
        """
        self.initialize_model()
        self.initialize_controller()
        self.initialize_view()

    @classmethod
    def get_instance(cls, key: str, factory: Callable[[str], IFacade]) -> IFacade:
        """
        Facade Multiton Factory method

        :param key: The key used to identify the Multiton instance
        :type key: str
        :param factory: A callable object that takes a string parameter and returns an instance of the IFacade interface
        :type factory: Callable[[str], IFacade]
        :return: the Multiton instance of the Facade
        """
        with cls.instanceMapLock:
            if key not in cls.instanceMap:
                cls.instanceMap[key] = factory(key)
        return cls.instanceMap.get(key)

    def initialize_controller(self):
        """
        Initialize the `Controller`.

        Called by the `initialize_facade` method.
        Override this method in your subclass of `Facade`
        if one or both of the following are true:

        You wish to initialize a different `IController`.
        You have `Commands` to register with the `Controller` at startup.

        If you don't want to initialize a different `IController`,
        call `super.initialize_controller()` at the beginning of your method, then register `Proxy`.

        Note: This method is rarely overridden; in practice, you are more
        likely to use a `Command` to create and register `Proxy`s
        with the `Model`, since `Proxy` with mutable data will likely
        need to send `INotification` and thus will likely want to fetch a reference to
        the `Facade` during their construction.

        :return: None
        """
        self.controller: IController = Controller.get_instance(self.multitonKey, lambda k: Controller(k))

    def initialize_model(self):
        """
        Initialize the `Model`.

        Called by the `initialize_facade` method.
        Override this method in your subclass of `Facade`
        if one or both of the following are true:

        You wish to initialize a different `IModel`.

        You have `Proxy` to register with the Model that does not
        retrieve a reference to the Facade at construction time.

        If you don't want to initialize a different `IModel`,
        call `super.initialize_model()` at the beginning of your
        method, then register `Proxy`.

        Note: This method is rarely overridden; in practice, you are more
        likely to use a `Command` to create and register `Proxy` with the
        `Model`, since `Proxy` with mutable data will likely need to send
        `INotification` and thus will likely want to fetch a reference to
        the `Facade` during their construction.

        :return: None
        """
        self.model: IModel = Model.get_instance(self.multitonKey, lambda k: Model(k))

    def initialize_view(self):
        """
        Initialize the `View`.

        Called by the `initialize_facade` method.
        Override this method in your subclass of `Facade`
        if one or both of the following are true:

        You wish to initialize a different `IView`.

        You have `Observers` to register with the `View`

        If you don't want to initialize a different `IView`,
        call `super.initialize_view()` at the beginning of your
        method, then register `IMediator` instances.

        Note: This method is rarely overridden; in practice, you are more
        likely to use a `Command` to create and register `Mediator` with the
        `View`, since `IMediator` instances will need to send
        `INotification` and thus will likely want to fetch a reference
        to the `Facade` during their construction.

        :return: None
        """
        self.view: IView = View.get_instance(self.multitonKey, lambda k: View(k))

    def register_command(self, notification_name: str, factory: Callable[[str], ICommand]):
        """
        Register an `ICommand` with the `Controller` by Notification name.

        :param notification_name: The name of the `INotification` to associate the `ICommand` with
        :type notification_name: str
        :param factory: A factory function that will be used to create instances of the `ICommand`.
        :type factory: Callable[[str], ICommand]
        :return: None
        """
        self.controller.register_command(notification_name, factory)

    def has_command(self, notification_name: str) -> bool:
        """
        Check if a Command is registered for a given Notification

        :param notification_name: The name of the `INotification`
        :type notification_name: str
        :return: `True` if a command is registered for the notification, `False` otherwise.
        """
        return self.controller.has_command(notification_name)

    def remove_command(self, notification_name: str):
        """
        Remove a previously registered `ICommand` to `INotification` mapping from the Controller.

        :param notification_name: The name of the `INotification` to remove the `ICommand` mapping for.
        :type notification_name: str
        :return: None
        """
        self.controller.remove_command(notification_name)

    def register_proxy(self, proxy: IProxy):
        """
        Register an `IProxy` with the `Model` by name.

        :param proxy: The `IProxy` instance to be registered with the `Model`.
        :type proxy: IProxy
        :return: None.
        """
        self.model.register_proxy(proxy)

    def retrieve_proxy(self, proxy_name: str) -> IProxy:
        """
        Retrieve an `IProxy` from the `Model` by name.

        :param proxy_name: The name of the proxy to be retrieved.
        :type proxy_name: str
        :return: the `IProxy` instance previously registered with the given `proxyName`.
        """
        return self.model.retrieve_proxy(proxy_name)

    def has_proxy(self, proxy_name: str) -> bool:
        """
        Check if a Proxy is registered

        :param proxy_name: The name of the `IProxy`
        :type proxy_name: str
        :return: True if the proxy exists, False otherwise.
        """
        return self.model.has_proxy(proxy_name)

    def remove_proxy(self, proxy_name: str) -> IProxy:
        """
        Remove an `IProxy` from the `Model` by name.

        :param proxy_name: The `IProxy` to remove from the `Model`.
        :type proxy_name: IProxy
        :return: the `IProxy` that was removed from the `Model`
        """
        return self.model.remove_proxy(proxy_name)

    def register_mediator(self, mediator: IMediator):
        """
        Register a `IMediator` with the `View`.

        :param mediator: A reference to the `IMediator`
        :type mediator: IMediator
        :return: None
        """
        self.view.register_mediator(mediator)

    def retrieve_mediator(self, mediator_name: str) -> IMediator:
        """
        Retrieve an `IMediator` from the `View`.

        :param mediator_name: The name of the `IMediator`
        :type mediator_name: str
        :return: the `IMediator` previously registered with the given `mediator_name`.
        :rtype: IMediator
        """
        return self.view.retrieve_mediator(mediator_name)

    def has_mediator(self, mediator_name: str) -> bool:
        """
        Check if a Mediator is registered or not

        :param mediator_name: The name of the `IMediator`
        :type mediator_name: str
        :return: True if the mediator with the specified name exists in the view, False otherwise.
        :rtype: bool
        """
        return self.view.has_mediator(mediator_name)

    def remove_mediator(self, mediator_name: str) -> IMediator:
        """
        Remove an `IMediator` from the `View`.

        :param mediator_name: Name of the `IMediator` to be removed.
        :type mediator_name: str
        :return: The `IMediator` that was removed from the `View`
        :rtype: IMediator
        """
        return self.view.remove_mediator(mediator_name)

    def send_notification(self, notification_name: str, body: Any = None, _type: str = None):
        """
        Create and send an `INotification`.

        Keeps us from having to construct new notification
        instances in our implementation code.

        :param notification_name: The name of the notification.
        :type notification_name: str
        :param body: The body of the notification (optional).
        :type body: Any
        :param _type: The type of the notification (optional).
        :type _type: str
        :return: None
        """
        self.notify_observers(Notification(notification_name, body, _type))

    def notify_observers(self, notification: INotification):
        """
        Notify `Observer`.

        This method is left public mostly for backward
        compatibility, and to allow you to send custom
        notification classes using the facade.

        Usually you should just call sendNotification
        and pass the parameters, never having to
        construct the notification yourself.

        :param notification: The `INotification` to have the `View` notify `Observers` of.
        :type notification: INotification
        :return: None
        """
        self.view.notify_observers(notification)

    def initialize_notifier(self, key: str):
        """
        Set the Multiton key for this facade instance.

        Not called directly, but instead from the constructor when getInstance
        is invoked. It is necessary to be public to implement
        INotifier.

        :param key: A string representing the key for the notifier.
        :type key: str
        :return: None
        """
        self.multitonKey = key

    @classmethod
    def has_core(cls, key: str) -> bool:
        """
        Check if a Core is registered or not

        :param key: The multiton key for the Core in question
        :type key: str
        :return: True if an instance with the specified key exists, False otherwise.
        """
        return cls.instanceMap.get(key) is not None

    @classmethod
    def remove_core(cls, key: str):
        """
        Remove a Core.

        Remove the Model, View, Controller and Facade instances for the given
        key.

        :param key: The key representing the core components to be removed.
        :type key: str
        :return: None
        """
        if cls.instanceMap.get(key) is None:
            return
        Model.remove_model(key)
        View.remove_view(key)
        Controller.remove_controller(key)
        del cls.instanceMap[key]



# MacroCommand.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from typing import List, Callable

from puremvc.interfaces import ICommand, INotification
from puremvc.patterns.facade import Notifier


class MacroCommand(Notifier, ICommand):
    """
    A base `ICommand` implementation that executes other `ICommand`.

    A `MacroCommand` maintains a list of `ICommand` Class references called `SubCommands`.

    When `execute` is called, the `MacroCommand`
    instantiates and calls `execute` on each of its `SubCommand` turn.
    Each `SubCommand` will be passed a reference to the original
    `INotification` that was passed to the `MacroCommand`'s
    `execute` method.

    Unlike `SimpleCommand`, your subclass
    should not override `execute`, but instead, should
    override the `initialize_macro_command` method,
    calling `add_subcommand` once for each `SubCommand`
    to be executed.

    See Also
    --------
    :class:`puremvc.core.Controller`
    :class:`puremvc.patterns.observer.Notification`
    :class:`puremvc.patterns.command.SimpleCommand`
    """

    def __init__(self):
        """
        MacroCommand Constructor.

        You should not need to define a constructor, instead, override the `initialize_macro_command` method.

        If your subclass does define a constructor, be sure to call `super().__init__()`.

        :return: None
        """
        super().__init__()
        self._subcommands: List[Callable[[], ICommand]] = []

    def initialize_macro_command(self):
        """
        Initialize the `MacroCommand`.

        In your subclass, override this method to initialize the `MacroCommand` `SubCommand` list with `ICommand`
        class references like this:

        Initialize MyMacroCommand::
        
            def initialize_macro_command(self):
                self.add_subcommand(lambda: FirstCommand())
                self.add_subcommand(lambda: SecondCommand())
                self.add_subcommand(lambda: ThirdCommand())

        Note that `SubCommand` may be any `ICommand` implementor,
        `MacroCommand` or `SimpleCommands` are both acceptable.

        :return: None
        """
        return

    def add_subcommand(self, factory: Callable[[], ICommand]):
        """
        Add a `SubCommand`.

        The `SubCommands` will be called in First In/First Out (FIFO) order.

        :param factory: A callable object that returns an instance of ICommand.
        :type factory: Callable[[], ICommand]
        :return: None
        """
        self._subcommands.append(factory)

    def execute(self, notification: INotification):
        """
        Execute this `MacroCommand`'s `SubCommands`.

        The `SubCommands` will be called in First In/First Out (FIFO)
        order.

        :param notification: The `INotification` object to be passed to each`SubCommand`.
        :type notification: INotification
        :return: None
        """
        self.initialize_macro_command()
        while self._subcommands:
            factory = self._subcommands.pop(0)
            command = factory()
            command.initialize_notifier(self.multitonKey)
            command.execute(notification)



# Observer.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from typing import Any, Callable

from puremvc.interfaces import IObserver, INotification


class Observer(IObserver):
    """
    A base `IObserver` implementation.

    An `Observer` is an object that encapsulates information
    about an interested object with a method that should
    be called when a particular `INotification` is broadcast.

    In PureMVC, the `Observer` class assumes these responsibilities:

    Encapsulate the notification (callback) method of the interested object.

    Encapsulate the notification context (this) of the interested object.

    Provide methods for setting the notification method and context.

    Provide a method for notifying the interested object.

    See Also
    --------
    :class:`puremvc.core.View`
    :class:`puremvc.patterns.observer.Notification`
    """

    def __init__(self, notify_method: Callable[[INotification], None] = None, notify_context: Any = None):
        """
        Constructor.

        The notification method on the interested object should take
        one parameter of type `INotification`

        :param notify_method: The notification method of the interested object
        :type notify_method: Callable[[INotification], None]
        :param notify_context: the notification context of the interested object
        :type notify_context: Any
        """
        self._notify_method = notify_method
        self._notify_context = notify_context

    @property
    def notify_method(self) -> Callable[[INotification], None]:
        """
        Get the notification method.

        :return: The notify method.
        :rtype: Callable[[INotification]]
        """
        return self._notify_method

    @notify_method.setter
    def notify_method(self, value: Callable[[INotification], None]):
        """
        Set the notification context.

        :param value: A callable function that takes an INotification object as its parameter.
        :type value: Callable[[INotification], None]
        """
        self._notify_method = value

    @property
    def notify_context(self) -> Any:
        """
        Get the notification context.

        :return: The notify context.
        :rtype: Any
        """
        return self._notify_context

    @notify_context.setter
    def notify_context(self, value: Any):
        """
        Set the notification context.

        :param value: The notification context (self) of the interested object.
        :type value: Any
        :return: None
        """
        self._notify_context = value

    def notify_observer(self, notification: INotification):
        """
        Notify the interested object.

        :param notification: The `INotification` to pass to the interested
        :type notification: INotification
        :return: None
        """
        if self._notify_method is not None:
            self._notify_method(notification)

    def compare_notify_context(self, obj: Any) -> bool:
        """
        Compare an object to the notification context.

        :param obj: The object to compare with the notify context.
        :type obj: Any
        :return: True if the given object is equal to the notify context, False otherwise.
        :rtype: bool
        """
        return obj == self._notify_context



# Notifier.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from typing import Any

from puremvc.interfaces import IFacade, INotifier
from puremvc.patterns.facade import Facade


class Notifier(INotifier):
    """
    A Base `INotifier` implementation.

    `MacroCommand, Command, Mediator` and `Proxy`
    all have a need to send `Notifications`.

    The `INotifier` interface provides a common method called
    `sendNotification` that relieves implementation code of
    the necessity to actually construct `Notifications`.

    The `Notifier` class, which all the above-mentioned classes
    extend, provides an initialized reference to the `Facade`
    Multiton, which is required for the convenience method
    for sending `Notifications`, but also eases implementation as these
    classes have frequent `Facade` interactions and usually require
    access to the facade anyway.

    NOTE: In the MultiCore version of the framework, there is one caveat to
    notifiers, they cannot send notifications or reach the facade until they
    have a valid multitonKey.

    The multitonKey is set:

    - on a Command when it is executed by the Controller
    - on a Mediator is registered with the View
    - on a Proxy is registered with the Model.

    See Also
    --------
    :class:`puremvc.patterns.proxy.Proxy`
    :class:`puremvc.patterns.facade.Facade`
    :class:`puremvc.patterns.mediator.Mediator`
    :class:`puremvc.patterns.command.MacroCommand`
    :class:`puremvc.patterns.command.SimpleCommand`
    """

    """Multiton error message"""
    MULTITON_MSG = "multitonKey for this Notifier not yet initialized!"

    def __init__(self):
        """Initialise the `INotifier` instance with an empty multiton key"""
        self.multitonKey = None

    def send_notification(self, notification_name: str, body: Any = None, note_type: str = None):
        """
        Create and send an `INotification`.

        Keeps us from having to construct new INotification instances in our
        implementation code.

        :param notification_name: The name of the notification to be sent.
        :type notification_name: str
        :param body: The body of the notification (optional). Default is None.
        :type body: Any, optional
        :param note_type: The type of the notification (optional). Default is None.
        :type note_type: str, optional
        :return: None
        """
        self.facade.send_notification(notification_name, body, note_type)

    def initialize_notifier(self, key: str):
        """
        Initialize this INotifier instance.

        This is how a Notifier gets its multitonKey. Calls to sendNotification
        or to access the facade will fail until after this method has been
        called.

        Mediators, Commands or Proxies may override this method in order to
        send notifications or access the Multiton Facade instance as soon as
        possible. They CANNOT access the facade in their constructors, since
        this method will not yet have been called.

        :param key: The multitonKey for this INotifier to use
        :type key: str
        :return: None
        """
        self.multitonKey = key

    @property
    def facade(self) -> IFacade:
        """
        Return the Multiton Facade instance

        :return: The instance of IFacade.
        :rtype: IFacade
        """
        if self.multitonKey is None:
            raise Exception(self.MULTITON_MSG)
        return Facade.get_instance(self.multitonKey, lambda key: Facade(key))



# IModel.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import ABC, abstractmethod

from .IProxy import IProxy


class IModel(ABC):
    """
    The interface definition for a PureMVC Model.

    In PureMVC, `IModel` implementors provide
    access to `IProxy` objects by named look up.

    An `IModel` assumes these responsibilities:

    Maintain a cache of `IProxy` instances
    Provide methods for registering, retrieving, and removing `IProxy` instances
    """

    @abstractmethod
    def register_proxy(self, proxy: IProxy):
        """
        Register an `IProxy` instance with the `Model`.

        :param proxy: An object reference to be held by the `Model`.
        :type proxy: IProxy
        """
        pass

    @abstractmethod
    def retrieve_proxy(self, proxy_name: str) -> IProxy:
        """
        Retrieve an `IProxy` instance from the Model.

        :param proxy_name: The name of the proxy to retrieve.
        :type proxy_name: str
        :return: The `IProxy` instance previously registered with the given `proxy_name`.
        :rtype: IProxy
        """
        pass

    @abstractmethod
    def remove_proxy(self, proxy_name: str) -> IProxy:
        """
        Remove an `IProxy` instance from the Model.

        :param proxy_name: The name of the proxy to remove.
        :type proxy_name: str
        :return: The `IProxy` instance that was removed from the Model.
        :rtype: IProxy
        """
        pass

    @abstractmethod
    def has_proxy(self, proxy_name: str) -> bool:
        """
        Check if a proxy with the given name is registered.

        :param proxy_name: The name of the proxy to check.
        :type proxy_name: str
        :return: True if the proxy is registered, False otherwise.
        :rtype: bool
        """
        pass



# INotifier.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import abstractmethod, ABC
from typing import Any


class INotifier(ABC):
    """
    The interface definition for a PureMVC Notifier.

    `MacroCommand, Command, Mediator` and `Proxy`
    all have a need to send `Notifications`.

    The `INotifier` interface provides a common method called
    `send_notification` that relieves implementation code of
    the necessity to actually construct `Notifications`.

    The `Notifier` class, which all the above-mentioned classes
    extend, also provides an initialized reference to the `Facade`
    Singleton, which is required for the convenience method
    for sending `Notifications`, but also eases implementation as these
    classes have frequent `Facade` interactions and usually require
    access to the facade anyway.

    See Also
    --------
    :class:`puremvc.interfaces.IFacade`
    :class:`puremvc.interfaces.INotification`
    """

    @abstractmethod
    def send_notification(self, notification_name: str, body: Any = None, _type: str = None):
        """
        Send a `INotification`.

        Convenience method to prevent having to construct new notification
        instances in our implementation code.

        @param notification_name: The name of the notification to send
        @type notification_name: str
        @param body: the body of the notification (optional)
        @type body: Any
        @param _type: the type of the notification (optional)
        @type _type: str
        """
        pass

    @abstractmethod
    def initialize_notifier(self, key: str):
        """
        Initialise this `INotifier` instance.

        This is how a `Notifier` gets its `multiton_key`. Calls to `send_notification`
        or to access the facade will fail until after this method has been
        called.

        @param key: The `multiton_key` for this `INotifier` to use
        """
        pass



# ICommand.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import abstractmethod

from .INotification import INotification
from .INotifier import INotifier


class ICommand(INotifier):
    """
    The interface definition for a PureMVC Command

    See Also
    --------
    :class:`puremvc.interfaces.INotification`
    """

    @abstractmethod
    def execute(self, notification: INotification):
        """
        Execute the `ICommand`'s logic to handle a given `INotification`.

        :param notification: An `INotification` to handle.
        :type notification: INotification
        :return: None
        """
        pass



# Mediator.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from typing import Any

from puremvc.interfaces import IMediator, INotification
from puremvc.patterns.facade import Notifier


class Mediator(Notifier, IMediator):
    """
    A base `IMediator` implementation.

    See Also
    --------
    :class:`puremvc.core.View`
    """

    """NAME (str): The name of the `Mediator`"""
    NAME = "Mediator"

    def __init__(self, mediator_name: str = None, view_component: Any = None):
        """
        Constructor

        Typically, a `Mediator` will be written to serve
        one specific control or group controls and so,
        will not have a need to be dynamically named.

        :param mediator_name: A string representing the name of the mediator. If not provided, it defaults to self.NAME.
        :type mediator_name: str
        :param view_component: Any object representing the view component. If not provided, it defaults to None.
        :type view_component: Any
        """
        super().__init__()
        self._mediator_name = self.NAME if mediator_name is None else mediator_name
        self._view_component = view_component

    @property
    def mediator_name(self) -> str:
        """
        Get the name of the `Mediator`.

        :return: The name of the mediator.
        :rtype: str
        """
        return self._mediator_name

    @property
    def view_component(self) -> Any:
        """
        Get the `Mediator` view component.

        :return: The view component.
        :rtype: Any
        """
        return self._view_component

    @view_component.setter
    def view_component(self, value: Any):
        """
        Set the `IMediator` view component.

        :param value: The view component
        :type value: Any
        :return: None
        """
        self._view_component = value

    def list_notification_interests(self) -> [str]:
        """
        List the `INotification` names this
        `Mediator` is interested in being notified of.

        :return: List the list of `INotification` names
        :rtype: List[str]
        """
        return []

    def handle_notification(self, notification: INotification):
        """
        Handle `INotification`.

        Typically, this will be handled in an if/else statement,
        with one 'comparison' entry per `INotification`
        the `Mediator` is interested in.

        :param notification: The notification to be handled.
        :type notification: INotification
        :return: None
        """
        return

    def on_register(self):
        """Called by the View when the Mediator is registered"""
        return

    def on_remove(self):
        """Called by the View when the Mediator is removed"""
        return



# IMediator.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import abstractmethod
from typing import Any

from .INotification import INotification
from .INotifier import INotifier


class IMediator(INotifier):
    """
    The base interface for all Mediator classes.

    In PureMVC, `IMediator` implementors assume these responsibilities:

    Implement a common method which returns a list of all `INotification`
    the `IMediator` has interest in implementing a notification callback method.
    Implement methods that are called when the IMediator is registered or removed from the View.
    
    Additionally, `IMediator` typically:

    Act as an intermediary between one or more view components such as text boxes or
    list controls, maintaining references and coordinating their behavior.

    In Flash-based apps, this is often the place where event listeners are
    added to view components, and their handlers implemented.

    Respond to and generate `INotifications`, interacting with of
    the rest of the PureMVC app.

    When an `IMediator` is registered with the `IView`,
    the `IView` will call the `IMediator`'s
    `list_notification_interests` method. The `IMediator` will
    return an `Array` of `INotification` names which
    it wishes to be notified about.

    The `IView` will then create an `Observer` object
    encapsulating that `IMediator`'s (`handle_notification`) method
    and register it as an Observer for each `INotification` name returned by
    `list_notification_interests`.

    See Also
    --------
    :class:`puremvc.interfaces.INotification`
    """

    @property
    @abstractmethod
    def mediator_name(self) -> str:
        """
        Get the `IMediator` instance name.

        :return: The `IMediator` instance name
        """
        pass

    @property
    @abstractmethod
    def view_component(self) -> Any:
        """
        Get the `IMediator`'s view component.

        :return: The view component.
        :rtype: Any
        """
        pass

    @view_component.setter
    def view_component(self, value: Any):
        """
        Set the `IMediator`'s view component.

        :param value: The view component
        :type value: Any
        :return: None
        """
        pass

    @abstractmethod
    def list_notification_interests(self) -> [str]:
        """
        List `INotification` interests.

        :return: A list containing strings representing the notification interests.
        :rtype: [str]
        """
        pass

    @abstractmethod
    def handle_notification(self, notification: INotification):
        """
        Handle an `INotification`.

        :param notification: The `INotification` to be handled
        :type notification: INotification
        :return: None
        """
        pass

    @abstractmethod
    def on_register(self):
        """
        Called by the View when the Mediator is registered
        :return: None
        """
        pass

    @abstractmethod
    def on_remove(self):
        """
        Called by the View when the Mediator is removed
        :return: None
        """
        pass



# IFacade.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import ABC, abstractmethod
from typing import Callable

from .ICommand import ICommand
from .IMediator import IMediator
from .INotification import INotification
from .INotifier import INotifier
from .IProxy import IProxy


class IFacade(INotifier, ABC):
    """
    The interface definition for a PureMVC Facade.

    The Facade Pattern suggests providing a single class to act as a central point of communication
    for a subsystem.

    In PureMVC, the Facade acts as an interface between the core MVC actors (Model, View, Controller) and
    the rest of your application.

    :class:`puremvc.interfaces.IModel`
    :class:`puremvc.interfaces.IView`
    :class:`puremvc.interfaces.IController`
    :class:`puremvc.interfaces.ICommand`
    :class:`puremvc.interfaces.INotification`
    """

    @abstractmethod
    def register_proxy(self, proxy: IProxy):
        """
        Register a `Proxy` with the `Model` by name.

        :param proxy: The `IProxy` to be registered with the `Model`.
        :type proxy: IProxy
        :return: None
        """
        pass

    @abstractmethod
    def retrieve_proxy(self, proxy_name: str) -> IProxy:
        """
        Retrieve a `IProxy` from the `Model` by name.

        :param proxy_name: The name of the `IProxy` instance to be retrieved.
        :type proxy_name: str
        :return: the `IProxy` previously registered by `proxyName` with the `Model`.
        """
        pass

    @abstractmethod
    def remove_proxy(self, proxy_name: str) -> IProxy:
        """
        Remove an `IProxy` instance from the `Model` by name.

        :param proxy_name: The `IProxy` to remove from the `Model`.
        :type proxy_name: str
        :return: the `IProxy` that was removed from the `Model`
        :rtype: IProxy
        """
        pass

    @abstractmethod
    def has_proxy(self, proxy_name: str) -> bool:
        """
        Check if a Proxy is registered

        :param proxy_name: The name of the proxy to check.
        :type proxy_name: str
        :return: True if the proxy is currently registered, False otherwise.
        :rtype: bool
        """
        pass

    @abstractmethod
    def register_command(self, notification_name: str, factory: Callable[[], ICommand]):
        """
        Register an `ICommand` with the `Controller`.

        :param notification_name: The name of the `INotification` to associate the `ICommand` with.
        :type notification_name: str
        :param factory: A callable factory function that creates an instance of `ICommand`.
        :type factory: Callable[[], ICommand]
        :return: None
        """
        pass

    @abstractmethod
    def has_command(self, notification_name: str) -> bool:
        """
        Check if a Command is registered for a given `Notification`.

        :param notification_name: The name of the notification to check.
        :type notification_name: str
        :return: True if a command exists for the given `notification_name`, False otherwise.
        """
        pass

    @abstractmethod
    def remove_command(self, notification_name: str):
        """
        Remove a previously registered `ICommand` to `INotification` mapping from the Controller.

        :param notification_name: The name of the `INotification` to remove the `ICommand` mapping for
        :type notification_name: str
        :return: None
        """
        pass

    @abstractmethod
    def register_mediator(self, mediator: IMediator):
        """
        Register an `IMediator` instance with the `View`.

        :param mediator: A reference to the `IMediator` instance
        :type mediator: IMediator
        :return: None
        """
        pass

    @abstractmethod
    def retrieve_mediator(self, mediator_name: str) -> IMediator:
        """
        Retrieve an `IMediator` instance from the `View`.

        :param mediator_name: The name of the `IMediator` instance to retrieve
        :type mediator_name: str
        :return: The mediator object found, of type IMediator.
        """
        pass

    @abstractmethod
    def has_mediator(self, mediator_name: str) -> bool:
        """
        Check if a Mediator is registered or not

        :param mediator_name: The name of the mediator to check for.
        :type mediator_name: str
        :return: True if the mediator exists, False otherwise.
        """
        pass

    @abstractmethod
    def remove_mediator(self, mediator_name: str) -> IMediator:
        """
        Remove a `IMediator` instance from the `View`.

        :param mediator_name: name of the `IMediator` instance to be removed.
        :type mediator_name: str
        :return: The `IMediator` instance previously registered with the given `mediator_name`.
        """
        pass

    @abstractmethod
    def notify_observers(self, notification: INotification):
        """
        Notify `Observer`

        :param notification: The `INotification` to have the `View` notify `Observers` of.
        :type notification: INotification
        :return: None
        """
        pass



# Model.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

import threading
from typing import Callable, Dict

from puremvc.interfaces import IModel, IProxy


class Model(IModel):
    """
    A Multiton `IModel` implementation.

    In PureMVC, the `Model` class provides access to model objects (Proxies)
    by named lookup.

    The `Model` assumes these responsibilities:

    - Maintain a cache of `IProxy` instances.
    - Provide methods for registering, retrieving, and removing `IProxy` instances.

    Your application must register `IProxy` instances with the `Model`.
    Typically, you use an `ICommand` to create and register `IProxy`
    instances once the `Facade` has initialized the Core actors.

    See Also
    --------
    :class:`puremvc.patterns.proxy.Proxy`
    :class:`puremvc.interfaces.IProxy`
    """
    instanceMap: Dict[str, IModel] = dict()
    instanceMapLock: threading.Lock = threading.Lock()

    """MULTITON_MSG (str): Multiton error message"""
    MULTITON_MSG = "Model multiton instance for this key is already constructed!"

    def __init__(self, key: str):
        """
        Constructor.

        This IModel implementation is a Multiton,
        so you should not call the constructor
        directly, but instead call the static Multiton
        Factory method Model.get_instance(multiton_key)

        :param key: A unique key for this instance of the Model.
        :raises Exception: If an instance with the given `key` already exists in the `instanceMap`.
        :type key: str
        """
        if Model.instanceMap.get(key) is not None:
            raise Exception(Model.MULTITON_MSG)
        self.multitonKey: str = key
        Model.instanceMap[key] = self
        self.proxyMap: Dict[str, IProxy] = dict()
        self.proxyMapLock = threading.Lock()
        self.initialize_model()

    def initialize_model(self):
        """
        Initialize the `Model` instance.

        Called automatically by the constructor, this is your opportunity to
        initialize the Multiton instance in your subclass without overriding
        the constructor.

        :return: None
        """
        return

    @classmethod
    def get_instance(cls, key: str, factory: Callable[[str], IModel]) -> IModel:
        """
        Multiton Factory method.

        :param key: A string representing the key used to retrieve the instance.
        :type key: str
        :param factory: A factory function used to create new instances of `IModel`.
        :type factory: Callable[[str], IModel]
        :return: An instance of `IModel` associated with the given key.
        """
        with cls.instanceMapLock:
            if key not in cls.instanceMap:
                cls.instanceMap[key] = factory(key)
        return cls.instanceMap.get(key)

    def register_proxy(self, proxy: IProxy):
        """
        Register an `IProxy` with the `Model`.

        :param proxy: An `IProxy` to be held by the `Model`.
        :type proxy: IProxy
        :return: None
        """
        proxy.initialize_notifier(self.multitonKey)
        self.proxyMap[proxy.proxy_name] = proxy
        proxy.on_register()

    def retrieve_proxy(self, proxy_name: str) -> IProxy:
        """
        Retrieve an `IProxy` from the `Model`.

        :param proxy_name: The name of the proxy.
        :type proxy_name: str
        :return: the `IProxy` instance previously registered with the given `proxyName`.
        :rtype: IProxy
        """
        return self.proxyMap.get(proxy_name)

    def has_proxy(self, proxy_name: str) -> bool:
        """
        Check if a Proxy is registered

        :param proxy_name: A string representing the name of the proxy to check.
        :type proxy_name: str
        :return: Returns True if the proxy exists in the proxy map, False otherwise.
        :rtype: bool
        """
        return self.proxyMap.get(proxy_name) is not None

    def remove_proxy(self, proxy_name: str) -> IProxy:
        """
        Remove an `IProxy` from the `Model`.

        :param proxy_name: The name of the `IProxy` to remove.
        :type proxy_name: str
        :return: the `IProxy` that was removed from the `Model`
        :rtype: IProxy
        """
        proxy = self.proxyMap.get(proxy_name)
        if proxy:
            del self.proxyMap[proxy_name]
            proxy.on_remove()
        return proxy

    @classmethod
    def remove_model(cls, key: str):
        """
        Remove an IModel instance

        :param key: multiton_key of IModel instance to remove
        :type key: str
        :return: None
        """
        with cls.instanceMapLock:
            del cls.instanceMap[key]



# Notification.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from typing import Any

from puremvc.interfaces import INotification


class Notification(INotification):
    """
    A base `INotification` implementation.

    PureMVC does not rely upon underlying event models such
    as the one provided with Flash, and ActionScript 3 does
    not have an inherent event model.

    The Observer Pattern as implemented within PureMVC exists
    to support event-driven communication between the
    application and the actors of the MVC triad.

    Notifications are not meant to be a replacement for Events
    in Flex/Flash/Apollo. Generally, `IMediator` implementors
    place event listeners on their view components, which they
    then handle in the usual way. This may lead to the broadcast of `Notification` to
    trigger `ICommand` or to communicate with other `IMediators`. `IProxy` and `ICommand`
    instances communicate with each other and `IMediator`s
    by broadcasting `INotification`.

    A key difference between Flash `Event` and PureMVC
    `Notification` is that `Event` follows the
    'Chain of Responsibility' pattern, 'bubbling' up the display hierarchy
    until some parent component handles the `Event`, while
    PureMVC `Notification` follows a 'Publish/Subscribe'
    pattern. PureMVC classes need not be related to each other in a
    parent/child relationship to communicate with one another
    using `Notification`.

    See Also
    --------
    :class:`puremvc.patterns.observer.Observer`
    """

    def __init__(self, name: str, body: Any = None, _type: str = None):
        """
        Constructor.

        :param name: name of the `Notification` instance. (required)
        :param body: the `Notification` body. (optional)
        :param _type: The type of the note. Defaults to None.
        """
        self._name = name
        self._body = body
        self._type = _type

    @property
    def name(self) -> str:
        """
        Get the name of the `Notification` instance.

        :return: The name of the `Notification` instance.
        :rtype: str
        """
        return self._name

    @property
    def body(self) -> Any:
        """
        This method returns the value of the `_body` attribute.

        :return: The value of the `_body` attribute.
        :rtype: Any
        """
        return self._body

    @body.setter
    def body(self, body: Any):
        """
        Set the body of the `Notification` instance.

        :param body: The new value for the body.
        :type body: Any
        """
        self._body = body

    @property
    def type(self) -> str:
        """
        Get the body of the `Notification` instance.

        :return: The type of the object.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, note_type: str):
        """
        Set the type of the `Notification` instance.
        """
        self._type = note_type

    def __repr__(self):
        """
        Get the string representation of the `Notification` instance.

        :return: The string representation of the `Notification` instance.
        """
        body = "None" if self.body is None else repr(self.body)
        _type = "None" if self.type is None else repr(self.type)
        msg = "Notification Name: " + self.name + "\nBody:" + body + "\nType:" + _type
        return msg



# IView.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import ABC, abstractmethod
from typing import Any

from .IMediator import IMediator
from .INotification import INotification
from .IObserver import IObserver


class IView(ABC):
    """
    The interface definition for a PureMVC View.

    In PureMVC, `IView` implementors assume these responsibilities:

    In PureMVC, the `View` class assumes these responsibilities:

    Maintain a cache of `IMediator` instances.
    Provide methods for registering, retrieving, and removing `IMediators`.
    Managing the observer lists for each `INotification` in the application.
    Providing a method for attaching `IObservers` to an `INotification`'s observer list.
    Providing a method for broadcasting an `INotification`.
    Notifying the `IObservers` of a given `INotification` when it broadcast.

    See Also
    --------
    :class:`puremvc.interfaces.IMediator`
    :class:`puremvc.interfaces.IObserver`
    :class:`puremvc.interfaces.INotification`
    """

    @abstractmethod
    def register_observer(self, notification_name: str, observer: IObserver):
        """
        Register an `IObserver` to be notified of `INotifications` with a
        given name.

        :param notification_name: The name of the `INotifications` to notify this `IObserver` of
        :type notification_name: str
        :param observer: The `IObserver` to register
        :type observer: IObserver
        """
        pass

    @abstractmethod
    def notify_observers(self, notification: INotification):
        """
        Notify the `IObservers` for a particular `INotification`.

        All previously attached `IObservers` for this `INotification`'s
        list are notified and are passed a reference to the `INotification` in
        the order in which they were registered.

        :param notification: The `INotification` to notify `IObservers` of.
        :type notification: INotification
        """
        pass

    @abstractmethod
    def remove_observer(self, notification_name: str, notify_context: Any):
        """
         Remove a group of observers from the observer list for a given Notification name.

         :param notification_name: Which observer list to remove from
         :type notification_name: str
         :param notify_context: Removed the observers with this object as their notify_context
         :type notify_context: Any
         """
        pass

    @abstractmethod
    def register_mediator(self, mediator: IMediator):
        """
          Register an `IMediator` instance with the `View`.

          Registers the `IMediator` so that it can be retrieved by name,
          and further interrogates the `IMediator` for its
          `INotification` interests.

          If the `IMediator` returns any `INotification`
          names to be notified about, an `Observer` is created encapsulating
          the `IMediator` instance's `handleNotification` method
          and registering it as an `Observer` for all `INotifications` the
          `IMediator` is interested in.

          :param mediator: A reference to the `IMediator` instance
          """
        pass

    @abstractmethod
    def retrieve_mediator(self, mediator_name: str) -> IMediator:
        """
        Retrieve an `IMediator` from the `View`.

        :param mediator_name: The name of the `IMediator` instance to retrieve.
        :type mediator_name: str
        :return: The `IMediator` instance previously registered with the given `mediatorName`.
        :rtype: IMediator
        """
        pass

    @abstractmethod
    def has_mediator(self, mediator_name: str) -> bool:
        """
        Check if a Mediator is registered or not.

        :param mediator_name: Name of the `IMediator`
        :type mediator_name: str
        :return: whether a Mediator is registered with the given `mediatorName`.
        :rtype: bool
        """
        pass

    @abstractmethod
    def remove_mediator(self, mediator_name: str) -> IMediator:
        """
         Remove an `IMediator` from the `View`.

         :param mediator_name: Name of the `IMediator` instance to be removed.
         :type mediator_name: str
         :return: the `IMediator` that was removed from the `View`
         :rtype: IMediator
         """
        pass



# INotification.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import ABC, abstractmethod
from typing import Any


class INotification(ABC):
    """
    The interface definition for a PureMVC Notification.

    PureMVC does not rely upon underlying event models such
    as the one provided with Flash, and ActionScript 3 does
    not have an inherent event model.

    The Observer Pattern as implemented within PureMVC exists
    to support event-driven communication between the
    application and the actors of the MVC triad.

    Notifications are not meant to be a replacement for Events
    in Flex/Flash/AIR. Generally, `IMediator` implementors
    place event listeners on their view components, which they
    then handle in the usual way. This may lead to the broadcast of `Notification` to
    trigger `ICommand` or to communicate with other `IMediator`, `IProxy` and `ICommand`
    instances communicate with each other and `IMediator` by broadcasting `INotification`.

    A key difference between Flash `Event` and PureMVC
    `Notification` is that `Event` follows the
    'Chain of Responsibility' pattern, 'bubbling' up the display hierarchy
    until some parent component handles the `Event`, while
    PureMVC `Notification` follows a 'Publish/Subscribe'
    pattern. PureMVC classes need not be related to each other in a
    parent/child relationship to communicate with one another
    using `Notification`.

    See Also
    --------
    :class:`puremvc.interfaces.IView`
    :class:`puremvc.interfaces.IObserver`
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Get the name of the `INotification` instance.
        No setter, it should be set by constructor only

        :return: The name of the object.
        """
        pass

    @property
    @abstractmethod
    def body(self) -> Any:
        """
        Get the body of the `INotification` instance

        :return: The body of the method.
        :rtype: Any
        """
        pass

    @body.setter
    @abstractmethod
    def body(self, body: Any):
        """
        Set the body of the `INotification` instance

        :param body: The new body value to set for the method
        :type body: Any
        :return: None
        """
        pass

    @property
    @abstractmethod
    def type(self) -> str:
        """
        Get the type of the `INotification` instance

        :return: The type of the object.
        :rtype: str
        """
        pass

    @type.setter
    @abstractmethod
    def type(self, _type: str):
        """
        Set the type of the `INotification` instance

        :param _type: The type of the note.
        :type _type: str
        :return: None
        """
        pass

    def __repr__(self):
        """
        Get the string representation of the `INotification` instance

        :return: The string representation of the object.
        :rtype: str
        """
        pass



# Controller.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

import threading
from typing import Dict, Callable

from puremvc.interfaces import IController, ICommand, IView, INotification
from puremvc.patterns.observer import Observer
from .View import View


class Controller(IController):
    """
    A Multiton `IController` implementation.

    In PureMVC, the `Controller` class follows the
    'Command and Controller' strategy, and assumes these
    responsibilities:

    Remembering which `ICommand`s
    are intended to handle which `INotifications`.

    Registering itself as an `IObserver` with
    the `View` for each `INotification`
    that it has an `ICommand` mapping for.

    Creating a new instance of the proper `ICommand`
    to handle a given `INotification` when notified by the `View`.

    Calling the `ICommand`'s `execute`
    method, passing in the `INotification`.

    Your application must register `ICommands` with the
    Controller.

    The simplest way is to subclass `Facade`,
    and use its `initialize_controller` method to add your
    registrations.

    See Also
    --------
    :class:`.View`
    :class:`puremvc.patterns.observer.Observer`
    :class:`puremvc.patterns.observer.Notification`
    :class:`puremvc.patterns.command.SimpleCommand`
    :class:`puremvc.patterns.command.MacroCommand`
    """
    instanceMap: Dict[str, IController] = dict()
    instanceMapLock: threading.Lock = threading.Lock()

    """MULTITON_MSG (str): Multiton error message"""
    MULTITON_MSG = "Controller multiton instance for this key is already constructed!"

    def __init__(self, key: str):
        """
        This `IController` implementation is a Multiton, so you should not
        call the constructor directly, but instead call the static Factory
        method, passing the unique key for this instance
        `Controller.get_instance(multitonKey, lambda k: Controller(k))`

        :param key: The unique key identifier for the Controller instance.
        :type key: str
        :raises Exception: If an instance with the given `key` already exists in the `instanceMap`.
        """
        if Controller.instanceMap.get(key) is not None:
            raise Exception(Controller.MULTITON_MSG)
        self.multitonKey: str = key
        Controller.instanceMap[key] = self
        self.commandMap: Dict[str, Callable[[], ICommand]] = dict()
        self.view = None
        self.initialize_controller()

    def initialize_controller(self):
        """
        Initialize the Multiton `Controller` instance.

        Called automatically by the constructor.

        Note that if you are using a subclass of `View` in your application,
        you should also subclass `Controller` and override the
        `initialize_controller` method in the following way::

            def initialize_controller(self):
                self.view = MyView.get_instance(self.multitonKey, lambda: key: MyView(key))

        :return: None
        """
        self.view: IView = View.get_instance(self.multitonKey, lambda key: View(key))

    @classmethod
    def get_instance(cls, key: str, factory: Callable[[str], IController]) -> IController:
        """
        `Controller` Multiton Factory method.
        
        :param key: The key used to identify the instance.
        :type key: str
        :param factory: A factory function that creates a new instance of IController based on the provided key.
        :type factory: Callable[[str], IController]
        :return: The instance of IController associated with the given key.
        """
        with cls.instanceMapLock:
            if key not in cls.instanceMap:
                cls.instanceMap[key] = factory(key)
        return cls.instanceMap.get(key)

    def register_command(self, notification_name: str, factory: Callable[[], ICommand]):
        """
        Register a particular `ICommand` class as the handler for a particular
        `INotification`.

        If an `ICommand` has already been registered to
        handle `INotification`s with this name, it is no longer
        used, the new `ICommand` is used instead.

        The Observer for the new ICommand is only created if this is the
        first time an ICommand has been registered for this Notification name.

        :param notification_name: The name of the notification.
        :param factory: Callable that returns an instance of ICommand.
        :return: None.
        """
        if self.commandMap.get(notification_name) is None:
            self.view.register_observer(notification_name, Observer(self.execute_command, self))
        self.commandMap[notification_name] = factory

    def execute_command(self, notification: INotification):
        """
        Executes the specified command based on the given notification.

        :param notification: The notification to be executed.
        :type notification: INotification
        :return: None
        """
        factory = self.commandMap.get(notification.name)
        if factory is None:
            return
        command = factory()
        command.initialize_notifier(self.multitonKey)
        command.execute(notification)

    def has_command(self, notification_name: str) -> bool:
        """
        Check if a Command is registered for a given Notification

        :param notification_name: The name of the notification to check in the `commandMap`.
        :type notification_name: str
        :return: True if the `notification_name` exists in the `commandMap`, False otherwise.
        :rtype: bool
        """
        return self.commandMap.get(notification_name) is not None

    def remove_command(self, notification_name: str):
        """
        Remove a previously registered `ICommand` to `INotification` mapping.

        :param notification_name: The name of the notification associated with the command to be removed.
        :type notification_name: str
        :return: None
        """
        if self.has_command(notification_name):
            self.view.remove_observer(notification_name, self)
            del self.commandMap[notification_name]

    @classmethod
    def remove_controller(cls, key: str):
        """
        Remove an IController instance

        :param key: The key to identify the controller instance to be removed.
        :type key: str
        :return: None
        """
        with cls.instanceMapLock:
            del cls.instanceMap[key]



# Proxy.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from typing import Any

from puremvc.interfaces import IProxy
from puremvc.patterns.facade import Notifier


class Proxy(IProxy, Notifier):
    """
    A base `IProxy` implementation.

    In PureMVC, `Proxy` classes are used to manage parts of the
    application's data model.

    A `Proxy` might simply manage a reference to a local data object,
    in which case interacting with it might involve setting and
    getting of its data in synchronous fashion.

    `Proxy` classes are also used to encapsulate the application's
    interaction with remote services to save or retrieve data, in which case,
    we adopt an asynchronous idiom; setting data (or calling a method) on the
    `Proxy` and listening for a `Notification` to be sent
    when the `Proxy` has retrieved the data from the service.

    See Also
    --------
    :class:`puremvc.core.Model`
    """

    """NAME (str): The name of the `Proxy`"""
    NAME = "Proxy"

    def __init__(self, proxy_name: str = None, data: Any = None):
        """
        Constructor

        :param proxy_name: The name of the proxy. Defaults to None.
        :type proxy_name: str
        :param data: The data associated with the object. Defaults to None.
        :type data: Any
        """
        super().__init__()
        self._proxy_name = self.NAME if proxy_name is None else proxy_name
        self.data = data

    @property
    def proxy_name(self) -> str:
        """
        Get the `Proxy` name

        :return: The proxy name.
        :rtype: str
        """
        return self._proxy_name

    @property
    def data(self) -> Any:
        """
        Get the `Proxy` data

        :return: The Proxy `data` object.
        :rtype: Any
        """
        return self._data

    @data.setter
    def data(self, value: Any):
        """
        Set the `Proxy` data

        :param value: The Proxy data object
        :type value: Any
        :return: None
        """
        self._data = value

    def on_register(self):
        """Called by the `Model` when the `Proxy` is registered"""
        return

    def on_remove(self):
        """Called by the `Model` when the `Proxy` is removed"""
        return



# View.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

import threading
from typing import Dict, List, Callable, Any

from puremvc.interfaces import IView, IMediator, IObserver, INotification
from puremvc.patterns.observer import Observer


class View(IView):
    """
    A Multiton `IView` implementation.

    In PureMVC, the `View` class assumes these responsibilities:

    Maintain a cache of `IMediator` instances.

    Provide methods for registering, retrieving, and removing `IMediators`.

    Notifying `IMediators` when they are registered or removed.

    Managing the observer lists for each `INotification` in the application.

    Providing a method for attaching `IObservers` to an `INotification`'s observer list.

    Providing a method for broadcasting an `INotification`.

    Notifying the `IObservers` of a given `INotification` when it broadcast.

    See Also
    --------
    :class:`puremvc.patterns.mediator.Mediator`
    :class:`puremvc.patterns.observer.Observer`
    :class:`puremvc.patterns.observer.Notification`
    """
    instanceMap: Dict[str, IView] = dict()
    instanceMapLock: threading.Lock = threading.Lock()

    """MULTITON_MSG (str): Multiton error message"""
    MULTITON_MSG = "View multiton instance for this key is already constructed!"

    def __init__(self, key: str):
        """
        Constructor.

        This `IView` implementation is a Multiton, so you should not call the
        constructor directly, but instead call the static Multiton Factory
        method `View.get_instance(multiton_key)`

        :param key: The unique key for this View instance.
        :type key: str
        :raises Exception: If an instance with the given `key` already exists in the `instanceMap`.
        """
        if View.instanceMap.get(key) is not None:
            raise Exception(View.MULTITON_MSG)
        self.multitonKey: str = key
        View.instanceMap[key] = self
        self.mediatorMap: Dict[str, IMediator] = dict()
        self.observerMap: Dict[str, List[IObserver]] = dict()
        self.initialize_view()

    @classmethod
    def get_instance(cls, key: str, factory: Callable[[str], IView]) -> IView:
        """
        View Multiton Factory method.

        :param key: The key associated with the desired instance.
        :param factory: A factory function that creates an instance of the desired class using the given key.
        :return: The instance associated with the given key.
        :rtype: IView
        """
        with cls.instanceMapLock:
            if key not in cls.instanceMap:
                cls.instanceMap[key] = factory(key)
        return cls.instanceMap.get(key)

    def initialize_view(self):
        """
        Initialize the Multiton `View` instance.

        Called automatically by the constructor, this is your opportunity to
        initialize the Multiton instance in your subclass without overriding
        the constructor.

        :return: None
        """
        return

    def register_observer(self, notification_name: str, observer: IObserver):
        """
        Register an `IObserver` to be notified of `INotifications` with a given name.

        :param notification_name: The name of the notification to register the observer for.
        :type notification_name: str
        :param observer: The observer object to register.
        :type observer: IObserver
        :return: None
        """
        if notification_name in self.observerMap:
            self.observerMap[notification_name].append(observer)
        else:
            self.observerMap[notification_name] = [observer]

    def notify_observers(self, notification: INotification):
        """
        Notify the `IObservers` for a particular `INotification`.

        All previously attached `IObservers` for this `INotification`'s
        list are notified and are passed a reference to the `INotification` in
        the order in which they were registered.

        :param notification: The notification to be sent to the observers.
        :type notification: INotification
        :return: None
        """
        observers = self.observerMap.get(notification.name)
        if observers:
            for observer in observers[:]:
                observer.notify_observer(notification)

    def remove_observer(self, notification_name: str, notify_context: Any):
        """
        Remove the observer for a given notify_context from an observer list for a given Notification name.

        :param notification_name: The name of the notification to remove the observer from.
        :type notification_name: str
        :param notify_context: The context object of the observer to remove.
        :type notify_context: Any
        :return: None
        """
        observers = self.observerMap.get(notification_name)

        for i, observer in enumerate(observers):
            if observer.compare_notify_context(notify_context):
                observers.pop(i)
                break

        if len(observers) == 0:
            del self.observerMap[notification_name]

    def register_mediator(self, mediator: IMediator):
        """
        Register an `IMediator` instance with the `View`.

        Registers the `IMediator` so that it can be retrieved by name,
        and further interrogates the `IMediator` for its
        `INotification` interests.

        If the `IMediator` returns any `INotification`
        names to be notified about, an `Observer` is created encapsulating
        the `IMediator` instance's `handleNotification` method
        and registering it as an `Observer` for all `INotifications` the
        `IMediator` is interested in.

        :param mediator: The mediator to register.
        :type mediator: IMediator
        :return: None
        """
        # Do not allow re-registration (you must to removeMediator first)
        if self.mediatorMap.get(mediator.mediator_name):
            return
        mediator.initialize_notifier(self.multitonKey)
        self.mediatorMap[mediator.mediator_name] = mediator

        interests = mediator.list_notification_interests()
        if interests:
            observer = Observer(mediator.handle_notification, mediator)
            for interest in interests:
                self.register_observer(interest, observer)
        mediator.on_register()

    def retrieve_mediator(self, mediator_name: str) -> IMediator:
        """
        Retrieve an `IMediator` from the `View`.

        :param mediator_name: The name of the mediator to retrieve.
        :type mediator_name: str
        :return: The mediator with the given name.
        :rtype: IMediator
        """
        return self.mediatorMap.get(mediator_name)

    def has_mediator(self, mediator_name: str) -> bool:
        """
        Check if a Mediator is registered or not

        :param mediator_name: The name of the mediator to check.
        :type mediator_name: str
        :return: Returns True if the mediator exists, False otherwise.
        :rtype: bool
        """
        return self.mediatorMap.get(mediator_name) is not None

    def remove_mediator(self, mediator_name: str) -> IMediator:
        """
        Remove an `IMediator` from the `View`.

        :param mediator_name: The name of the mediator to be removed.
        :type mediator_name: str
        :return: The removed mediator instance.
        :rtype: IMediator
        """
        mediator = self.mediatorMap.get(mediator_name)
        if mediator:
            for interest in mediator.list_notification_interests():
                self.remove_observer(interest, mediator)

            del self.mediatorMap[mediator_name]
            mediator.on_remove()

        return mediator

    @classmethod
    def remove_view(cls, key: str):
        """
        Remove an `IView` instance

        :param key: The key of the view to be removed.
        :return: None
        """
        with cls.instanceMapLock:
            del cls.instanceMap[key]



# IObserver.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from abc import ABC, abstractmethod
from typing import Any, Callable

from .INotification import INotification


class IObserver(ABC):
    """
    The interface definition for a PureMVC Observer.

    In PureMVC, `IObserver` implementors assume these responsibilities:

    Encapsulate the notification (callback) method of the interested object.

    Encapsulate the notification context of the interested object.

    Provide methods for setting the interested object notification method and context.

    Provide a method for notifying the interested object.

    PureMVC does not rely upon underlying event
    models such as the one provided with Flash,
    and ActionScript 3 does not have an inherent
    event model.

    The Observer Pattern as implemented within
    PureMVC exists to support event driven communication
    between the application and the actors of the
    MVC triad.

    An Observer is an object that encapsulates information
    about an interested object with a notification method that
    should be called when an `INotification` is broadcast. The Observer then
    acts as a proxy for notifying the interested object.

    Observers can receive `Notification`s by having their
    `notify_observer` method invoked, passing
    in an object implementing the `INotification` interface, such
    as a subclass of `Notification`.

    See Also
    --------
    :class:`puremvc.interfaces.IView`
    :class:`puremvc.interfaces.INotification`

    """

    @property
    @abstractmethod
    def notify_method(self) -> Callable[[INotification], None]:
        """
        Set the notification method.

        The notification method should take one parameter of type `INotification`

        :return: The notification (callback) method of the interested object
        :rtype: Callable[[INotification], None]
        """
        pass

    @notify_method.setter
    @abstractmethod
    def notify_method(self, value: Callable[[INotification], None]):
        """
        Set the notification context.

        :param value: The notification context (self) of the interested
        :type value: Callable[[INotification, None]
        """
        pass

    @property
    @abstractmethod
    def notify_context(self) -> Any:
        """
        Get the notify_context.

        :return: notify context
        :rtype: Any
        """
        pass

    @notify_context.setter
    @abstractmethod
    def notify_context(self, value: Any):
        """
        Sets the value of the `notify_context` attribute.

        :param value: The value to be set for the `notify_context` attribute.
        :type value: Any
        :return: None
        """
        pass

    @abstractmethod
    def notify_observer(self, notification: INotification):
        """
        Notify the interested object.

        :param notification: The `INotification` to pass to the interested object's notification method
        :type notification: INotification
        """
        pass

    @abstractmethod
    def compare_notify_context(self, obj: Any) -> bool:
        """
        Compare the given object to the notification context object.

        :param obj: The object to compare.
        :type obj: Any
        :return: boolean indicating if the notification context and the object are the same.
        """
        pass



from .IController import IController
from .IModel import IModel
from .IView import IView
from .ICommand import ICommand
from .IFacade import IFacade
from .IMediator import IMediator
from .INotification import INotification
from .INotifier import INotifier
from .IObserver import IObserver
from .IProxy import IProxy



# SimpleCommand.py
# PureMVC Python Multicore

# Copyright(c) 2025 Saad Shams <saad.shams@puremvc.org>
# Your reuse is governed by the BSD 3-Clause License

from puremvc.interfaces import INotification, ICommand

from puremvc.patterns.facade import Notifier


class SimpleCommand(Notifier, ICommand):
    """
    A base `ICommand` implementation.

    Your subclass should override the `execute`
    method where your business logic will handle the `INotification`.

    See Also
    --------
    :class:`puremvc.core.Controller`
    :class:`puremvc.patterns.observer.Notification`
    :class:`puremvc.patterns.command.MacroCommand`
    """

    def execute(self, notification: INotification):
        """
        Fulfill the use-case initiated by the given `INotification`.

        In the Command Pattern, an application use-case typically
        begins with some user action, which results in an `INotification` being broadcast, which
        is handled by business logic in the `execute` method of an `ICommand`.

        :param notification: The `INotification` to handle.
        :type notification: INotification
        :return: None
        """
        return
